"""Utilities for working with awaitable objects returned by async methods."""

from __future__ import annotations

import asyncio
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generator

import rich.repr
from textual.message_pump import MessagePump

Runnable = Awaitable | Callable[[], Awaitable | Any] | None


@rich.repr.auto(angular=True)
class AwaitCompleteOrNoop:
    """An 'optionally-awaitable' object which runs coroutines (or other awaitables) concurrently, or exits immediately.

    Functionally equivalent to `textual.await_complete.AwaitComplete`, but can be instantiated without an async loop
    if there is nothing to await. This ensures that the result provides a consistent interface, whether there
    are async tasks to complete or not.

    If there are awaitables, this must be called from within an async loop.
    """

    def __init__(self, *awaitables: Awaitable) -> None:
        """Initialize an AwaitCompleteOrNoop.

        Args:
            awaitables: Optional awaitables to run concurrently. If none are provided, this will act as a no-op.
        """
        if awaitables:
            self._future = gather(*awaitables)
        else:
            self._future = None

    def __await__(self) -> Generator[Any, None, Any]:
        """Wait for the tasks to complete, or return immediately if no tasks."""
        if not self._future:
            yield None
            return None
        return self._future.__await__()

    async def __call__(self) -> Any:
        """Wait for the tasks to complete, or return immediately if no tasks."""
        return await self

    def call_next(self, node: MessagePump) -> AwaitCompleteOrNoop:
        """Await after the next message.

        Args:
            node: The node which created the object.
        """
        node.call_next(self)
        return self

    @property
    def is_done(self) -> bool:
        """True if the task has completed or there was no task."""
        if not self._future:
            return True
        return self._future.done()

    @property
    def exception(self) -> BaseException | None:
        """The exception raised if the awaitables failed, or None if all succeeded (or no task)."""
        if not self._future:
            return None
        return self._future.exception() if self.is_done else None


def gather(*aws: Runnable, return_exceptions: bool = True) -> asyncio.Future:
    """Return a future aggregating results from the given functions/coroutines/futures.

    Args:
        aws: Awaitables, or functions to wrap in awaitables, to run concurrently.
        return_exceptions: Whether to raise exceptions, or return as results.

    Returns:
        A future aggregating all the results from the given functions/coroutines/futures.
    """
    results = []
    for awaitable in aws:
        if asyncio.iscoroutinefunction(awaitable):
            results.append(awaitable())
        elif asyncio.iscoroutine(awaitable):
            results.append(awaitable)
        else:
            results.append(_make_awaitable(awaitable)())
    return asyncio.gather(*results, return_exceptions=return_exceptions)


def _make_awaitable(func: Callable) -> Any:
    """Create an awaitable function from a non-async function."""

    async def _awaitable() -> Any:
        return func()

    return _awaitable



"""Extended Textual DataTable widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets
from typing_extensions import Literal

from .._extensions import Callbacks
from .._extensions import WidgetExtension


class DataTable(widgets.DataTable, WidgetExtension):  # pylint: disable=too-many-ancestors
    """An extended tabular widget that contains data."""

    def __init__(  # pylint: disable=too-many-locals,too-many-arguments
        self,
        *,
        show_header: bool = True,
        show_row_labels: bool = True,
        fixed_rows: int = 0,
        fixed_columns: int = 0,
        zebra_stripes: bool = False,
        header_height: int = 1,
        show_cursor: bool = True,
        cursor_foreground_priority: Literal["renderable", "css"] = "css",
        cursor_background_priority: Literal["renderable", "css"] = "renderable",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize the data table widget.

        Args:
            show_header: Whether to show the table header.
            show_row_labels: Whether to show labels on every row.
            fixed_rows: Fixed number of rows to display.
            fixed_columns: Fixed number of columns to display
            zebra_stripes: Whether to display the rows with alternating background highlight style.
            show_cursor: Whether to highlight the currently active cursor highlight.
            header_height: Height the header row.
            cursor_foreground_priority: Whether to prioritize the cursor CSS foreground or the renderable foreground.
            cursor_background_priority: Whether to prioritize the cursor CSS background or the renderable background.
            name: The name of the content switcher.
            id: The ID of the content switcher in the DOM.
            classes: The CSS classes of the content switcher.
            disabled: Whether the content switcher is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            show_header=show_header,
            show_row_labels=show_row_labels,
            fixed_rows=fixed_rows,
            fixed_columns=fixed_columns,
            zebra_stripes=zebra_stripes,
            header_height=header_height,
            show_cursor=show_cursor,
            cursor_foreground_priority=cursor_foreground_priority,
            cursor_background_priority=cursor_background_priority,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual radio button widget."""

from textual import widgets

from .._extensions import ToggleButtonFactory


class RadioButton(ToggleButtonFactory(widgets.RadioButton)):  # pylint: disable=too-many-ancestors
    """An extended radio button widget that represents a boolean value."""



"""Extended Textual label widget for displaying text-oriented renderables."""

from textual import widgets

from .._extensions import StaticFactory


class Label(StaticFactory(widgets.Label)):
    """An extended, simple, label widget for displaying text-oriented renderables."""



"""Extended Textual RichLog widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets

from .._extensions import Callbacks
from .._extensions import WidgetExtension


class RichLog(widgets.RichLog, WidgetExtension):
    """An extended widget for logging text."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        max_lines: int | None = None,
        min_width: int = 78,
        wrap: bool = False,
        highlight: bool = False,
        markup: bool = False,
        auto_scroll: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize a RichLog widget.

        Args:
            max_lines: Maximum number of lines in the log or `None` for no maximum.
            min_width: Minimum width of renderables.
            wrap: Enable word wrapping (default is off).
            highlight: Automatically highlight content.
            markup: Apply Rich console markup.
            auto_scroll: Enable automatic scrolling to end.
            name: The name of the text log.
            id: The ID of the text log in the DOM.
            classes: The CSS classes of the text log.
            disabled: Whether the text log is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            max_lines=max_lines,
            min_width=min_width,
            wrap=wrap,
            highlight=highlight,
            markup=markup,
            auto_scroll=auto_scroll,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual Pretty widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets

from .._extensions import Callbacks
from .._extensions import WidgetExtension


class Pretty(widgets.Pretty, WidgetExtension):
    """An extended pretty-printing widget."""

    DEFAULT_CSS = """
    Pretty {
        height: auto;
    }
    """

    def __init__(
        self,
        obj: Any,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize the Pretty widget.

        Args:
            obj: The object to pretty-print.
            name: The name of the pretty widget.
            id: The ID of the pretty in the DOM.
            classes: The CSS classes of the pretty.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            object=obj,
            name=name,
            id=id,
            classes=classes,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Utilities for running tasks in parallel."""

from __future__ import annotations

import multiprocessing
import threading
from concurrent import futures
from multiprocessing.context import BaseContext
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable


class ParallelPoolExecutor:
    """A pool of threads or processes to run asynchronous tasks in parallel.

    Closely matches `concurrent.futures.Executor` interfaces for greatest parity. Additional logic added
    to control indexing, ordering, multiprocess context handling, and cleanup.

    This pool is safe to use with Textual while an application is showing. Compared to `multiprocessing.pool.ThreadPool`
    and `multiprocessing.pool.Pool`, this should not encounter errors with file descriptors.

    Example:
        import urllib.request

        URLS = [
            'https://docs.python.org/3.9/library/',
            'https://docs.python.org/3.10/library/',
            'https://docs.python.org/3.11/library/',
            'https://docs.python.org/3.12/library/',
        ]

        def load_url(url: str, timeout: int) -> bytes:
            with urllib.request.urlopen(url, timeout=timeout) as conn:
                return conn.read()

        with ParallelPoolExecutor(max_workers=3) as pool:
            pool.submit(
                funcs=[load_url for url in URLS],
                func_args=[(url, 60) for url in URLS],
            )
            for result in pool.as_completed():
                print(len(result))
    """

    def __init__(
        self,
        max_workers: int | None = None,
        use_threads: bool = True,
        thread_name_prefix: str = "",
        mp_context: BaseContext | str | None = None,
        initializer: Callable[[], None] | None = None,
        initargs: tuple = (),
        clear_results: bool = True,
    ) -> None:
        """Create an executor pool of threads or processes that can execute calls in parallel.

        Args:
            max_workers: The maximum amount of threads/processes that can be used to execute the calls.
                See `concurrent.futures.ThreadPoolExecutor` and `concurrent.futures.ProcessPoolExecutor` for defaults.
            use_threads: Use multithreading instead of multiprocessing to improve resource management.
                Threads are recommended for I/O bound tasks, or CPU bound tasks which release the GIL.
            thread_name_prefix: An optional name prefix to give threads.
            mp_context: A multiprocessing context, or name, to launch the workers if threading is disabled.
                e.g. 'fork', 'forkserver', and 'spawn'.
            initializer: A callable used to initialize worker threads/processes.
            initargs: A tuple of arguments to pass to the initializer.
            clear_results: Whether to dereference future results after yield to allow faster garbage collection.
        """
        self._clear_results = clear_results
        self._completed = 0
        self._submitted = 0
        self._futures: dict[int, futures.Future] = {}
        self._ready: dict[int, Any] = {}

        if use_threads:
            self._executor = futures.ThreadPoolExecutor(  # pylint: disable=consider-using-with
                max_workers=max_workers,
                thread_name_prefix=thread_name_prefix,
                initializer=initializer,
                initargs=initargs,
            )
        else:
            mp_context = multiprocessing.get_context(mp_context) if isinstance(mp_context, str) else mp_context
            self._executor = futures.ProcessPoolExecutor(  # pylint: disable=consider-using-with
                max_workers=max_workers,
                mp_context=mp_context,
                initializer=initializer,
                initargs=initargs,
            )

    def __enter__(self) -> ParallelPoolExecutor:
        """Create an executor pool that will automatically release resources on exit."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> bool:
        """Release the executor pool resources."""
        self.shutdown()
        return False

    def as_completed(
        self,
        ordered: bool = False,
        with_index: bool = False,
        exit_on_error: bool = True,
    ) -> Generator[Any | tuple[int, Any], None, None]:
        """Iterate over results as each pending Future completes.

        Args:
            ordered: Return completed results in the same order requested.
                Pending results remain in memory until yielded.
            with_index: Return the index of the result with the result.
            exit_on_error: Whether to raise and abort on exceptions, or pass the exceptions as results.

        Yields:
            Result of a completed future if indexing is disabled, or index and result if enabled.
        """
        if self._completed == self._submitted:
            self._executor.shutdown()
            return
        if len(self._workers) == 0:
            raise futures.BrokenExecutor("All workers in pool terminated prematurely")

        for future in futures.as_completed(self._futures.values()):
            future_index, future_result = future.result()
            self._ready[future_index] = future_result
            self._futures.pop(future_index)
            if self._clear_results:
                # Dereference the future result value after it was yielded to allow garbage collection ASAP.
                future._result = None  # pylint: disable=protected-access
            for index, result in sorted(self._ready.items()) if ordered else list(self._ready.items()):
                if ordered and index != self._completed:
                    continue
                self._ready.pop(index)
                self._completed += 1
                if isinstance(result, BaseException) and exit_on_error:
                    raise result
                if with_index:
                    yield (index, result) if with_index else result
                else:
                    yield result

    @staticmethod  # Use static method to prevent pickling issues with subprocessing pools.
    def _call_with_index(
        index: int,
        func: Callable,
        args: Iterable,
        kwargs: dict[str, Any],
    ) -> tuple[int, Any]:
        """Run a function, capture failures, and return the index representing the future."""
        try:
            result = func(*args, **kwargs)
        except BaseException as error:  # pylint: disable=broad-except
            result = error
        return index, result

    @property
    def completed(self) -> int:
        """Number of tasks completed by the pool over its lifetime."""
        return self._completed

    def shutdown(self, wait: bool = False, *, cancel_futures: bool = True) -> None:
        """Stop the executor pool and release resources.

        This method is safe to call several times. No other methods can be called after this one.
        Processes will exit immediately. Threads will run until their active tasks complete.
        If interrupting threads is needed, the functions submitted must provide their own interruptions.

        Args:
            wait: Wait until all running futures have finished, and the resources used by the executor are reclaimed.
            cancel_futures: Cancel all pending futures. Futures that are completed or running will not be cancelled.
        """
        # Snapshot active workers before shutdown to ensure cleanup can be performed.
        # Shutdown will clear the active workers, even if they are still running.
        workers = self._workers
        try:
            self._executor.shutdown(wait=wait, cancel_futures=cancel_futures)
        finally:
            self._futures.clear()
            self._ready.clear()
            if isinstance(self._executor, futures.ProcessPoolExecutor):
                self._stop_processes(workers)

    @staticmethod
    def _stop_processes(workers: dict[int, multiprocessing.Process]) -> None:
        """Stop any running processes used by the executor pool.

        Do not call directly. Should only be called by executor pool on shutdown.

        Mimics `multiprocessing.Pool` to attempt graceful, immediate, shutdowns.
        Sends termination signals, waits (joins), and finally kills, if processes fail to exit successfully.
        Additional cleanup will be attempted by python before exit if any fail to stop.
        """
        try:
            for process in workers.values():
                if process.exitcode is None:
                    process.terminate()
        finally:
            try:
                for process in workers.values():
                    if process.is_alive():
                        process.join()
            finally:
                for process in workers.values():
                    if process.is_alive():
                        process.kill()

    def submit(
        self,
        funcs: Callable | list[Callable],
        func_args: tuple | list[tuple] | None = None,
        func_kwargs: dict | list[dict] | None = None,
    ) -> list[futures.Future]:
        """Submit callables to be executed with the given arguments.

        Schedules the callables to be executed as func(*args, **kwargs), and returns Future instances representing
        the execution of the callables.

        Args:
            funcs: One or more callables to be submitted to the pool to be run by the executor workers.
                Execution will begin immediately if any workers are idle.
            func_args: Positional arguments to send to the functions.
                Number of argument sets must match number of functions. Use None if no positional arguments are used.
            func_kwargs: Keyword arguments to send to the functions.
                Number of argument sets must match number of functions. Use None if no keyword arguments are used.

        Returns:
            One Future for each call.
        """
        if not isinstance(funcs, list):
            funcs = [funcs]
        if func_args is not None and not isinstance(func_args, list):
            func_args = [func_args]
        if func_kwargs is not None and not isinstance(func_kwargs, list):
            func_kwargs = [func_kwargs]

        # Ensure there is an arg and kwarg set for every function. Arguments are optional, but the amount provided must
        # match function length in order to be zipped, otherwise functions could be skipped.
        if not func_args:
            func_args = [() for _ in funcs]
        if len(funcs) != len(func_args):
            raise ValueError(
                f"Length of positional argument sets does not match number of functions: {len(func_args)}/{len(funcs)}"
            )
        if not func_kwargs:
            func_kwargs = [{} for _ in funcs]
        if len(funcs) != len(func_kwargs):
            raise ValueError(
                f"Length of keyword argument sets does not match number of functions: {len(func_kwargs)}/{len(funcs)}"
            )

        new_futures = []
        for index, (func, args, kwargs) in enumerate(zip(funcs, func_args, func_kwargs)):
            index += self._submitted
            future = self._executor.submit(self._call_with_index, index, func, args, kwargs)
            new_futures.append(future)
            self._futures[index] = future
        self._submitted += len(funcs)
        return new_futures

    @property
    def submitted(self) -> int:
        """Number of tasks submitted to the pool over its lifetime."""
        return self._submitted

    @property
    def _workers(self) -> dict[int, multiprocessing.Process] | set[threading.Thread]:
        """The active executor pool workers."""
        if isinstance(self._executor, futures.ThreadPoolExecutor):
            workers = self._executor._threads or set()  # pylint: disable=protected-access
        else:
            workers = self._executor._processes or {}  # pylint: disable=protected-access
        return workers


def parallelize(
    funcs: list[Callable],
    func_args: list[tuple] | None = None,
    func_kwargs: list[dict] | None = None,
    max_workers: int = None,
    use_threads: bool = True,
    mp_context: BaseContext | str | None = "fork",
    with_index: bool = False,
    ordered: bool = False,
    exit_on_error: bool = True,
) -> Generator[Any | tuple[int, Any] | Any, None, None]:
    """Run functions using a parallel processing pool.

    Args:
        funcs: Functions to run in parallel.
        func_args: Positional arguments to send to the functions.
            Number of argument sets must match number of functions. Use None if no positional arguments are used.
        func_kwargs: Keyword arguments to send to the functions.
            Number of argument sets must match number of functions. Use None if no keyword arguments are used.
        max_workers: Maximum number of workers to use to execute the functions.
        use_threads: Use multithreading instead of multiprocessing to improve resource management.
            Threads are recommended for I/O bound tasks, or CPU bound tasks which release the GIL.
        mp_context: A multiprocessing context, or name, to launch the workers if threading is disabled.
            e.g. 'fork', 'forkserver', and 'spawn'.
        ordered: Return completed results in the same order requested. Other results wait in memory until yielded.
        with_index: Return the index of the result with the result. Allows for tracking without forcing ordered.
        exit_on_error: Whether to raise and abort on exceptions, or pass the exceptions as results.

    Yields:
        Result of a completed future if indexing is disabled, or index and result if enabled.
    """
    with ParallelPoolExecutor(
        max_workers=max_workers,
        use_threads=use_threads,
        mp_context=mp_context,
    ) as pool:
        pool.submit(funcs, func_args=func_args, func_kwargs=func_kwargs)
        for result in pool.as_completed(ordered=ordered, with_index=with_index, exit_on_error=exit_on_error):
            yield result



"""Extended Textual Tree widgets."""

from typing import Any
from typing import Iterable

from rich.highlighter import ReprHighlighter
from rich.text import Text
from rich.text import TextType
from textual import events
from textual import widgets
from textual.widgets.tree import TreeDataType
from textual.widgets.tree import TreeNode

from . import Tree
from ._extensions import Callbacks


class LazyTree(Tree):  # pylint: disable=too-many-ancestors
    """Widget to lazily render an interactive tree from a dataset.

    Content will be loaded as expanded to maximize performance.
    """

    def __init__(
        self,
        label: TextType,
        data: TreeDataType | None = None,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize a Tree that will lazily render the data.

        Args:
            label: The label of the root node of the tree.
            data: The optional data to associate with the root node of the tree.
            name: The name of the Tree.
            id: The ID of the tree in the DOM.
            classes: The CSS classes of the tree.
            disabled: Whether the tree is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            label=label,
            data=data,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self._highlighter = ReprHighlighter()
        self._initialized = {}

    def _initialize_node(self, node: TreeNode | None = None, recursive: bool = False) -> None:
        """Add contents for a node."""
        node = node or self.cursor_node
        if not self._initialized.get(id(node)) and node.allow_expand:
            data = node.data
            if isinstance(data, dict):
                for key, value in data.items():
                    label, allow_expand = self._get_node_config(key, value)
                    node.add(label, value, allow_expand=allow_expand)
            elif isinstance(data, list):
                for index, value in enumerate(data):
                    label, allow_expand = self._get_node_config(str(index), value)
                    node.add(label, value, allow_expand=allow_expand)
            self._initialized[id(node)] = True
        if recursive:
            for child in node.children:
                self._initialize_node(child, recursive=True)

    def _get_node_config(self, name: str, value: Any) -> tuple[Text, bool]:
        """Provide the configuration for a new node, such as the label and whether it can be expanded."""
        if isinstance(value, dict):
            allow_expand = True
            label = Text(f"{name}={{}}") if name else Text("{}")
        elif isinstance(value, list):
            allow_expand = True
            label = Text(f"{name}=[]") if name else Text("[]")
        else:
            allow_expand = False
            if name:
                label = Text.assemble(Text.from_markup(f"[b]{name}[/b]="), self._highlighter(repr(value)))
            else:
                label = Text(repr(value))
        return label, allow_expand

    async def _on_tree_node_expanded(self, event: widgets.Tree.NodeExpanded) -> None:
        """Lazily add the contents of the node after it is expanded."""
        self._initialize_node(event.node)

    def collapse_all(
        self,
        node: TreeNode | None = None,
    ) -> None:
        """Collapse a node and all the nodes below it.

        Args:
            node: The node to collapse. Defaults to selected node.
        """
        node = node or self.cursor_node
        if node:
            disabled = Tree.NodeCollapsed in self._disabled_messages
            if not disabled:
                self.disable_messages(Tree.NodeCollapsed)
            with self.app.batch_update():
                node.collapse_all()
            if not disabled:
                self.enable_messages(Tree.NodeCollapsed)

    def expand_all(
        self,
        node: TreeNode | None = None,
    ) -> None:
        """Expand a node and all the nodes below it.

        Lazy expansion is bypassed, immediately populating the entire branch.

        Args:
            node: The node to collapse. Defaults to selected node.
        """
        node = node or self.cursor_node
        if node:
            self._initialize_node(node, recursive=True)
            disabled = Tree.NodeExpanded in self._disabled_messages
            if not disabled:
                self.disable_messages(Tree.NodeExpanded)
            with self.app.batch_update():
                node.expand_all()
            if not disabled:
                self.enable_messages(Tree.NodeExpanded)



"""Extended Textual widget that is an item within a ListView, and contains metadata about the selection."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets
from textual.widget import Widget

from ._extensions import Callbacks
from ._extensions import WidgetExtension
from ._textual._label import Label


class ListItem(widgets.ListItem, WidgetExtension):
    """An extended widget that is an item within a ListView, and contains metadata about the selection."""

    # Recommended events to ignore when widgets are used in ListViews to prevent large unneeded event batches.
    default_disabled_messages = (
        events.Mount,
        events.Unmount,
        events.Show,
        events.Hide,
        events.Resize,
    )

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        data: Any = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize a ListItem with extension arguments.

        Args:
            *children: Child widgets.
            name: The name of the widget.
                If no child is provided for display, and no "data" with "label" key, name will be used in a Label.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            data: Optional data associated with the list item.
                If no child is provided for display, the data will be searched for a "label" key to use in a Label.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        if not children:
            label = name
            if data and "label" in data:
                label = data["label"]
            if label:
                children = [Label(label, disabled_messages=ListItem.default_disabled_messages)]

        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled)
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self.data = data
        self.menu_index: int | None = None



"""Collection of new, and extended, Textual widgets.

"Extended" Textual widgets strive to maintain 100% parity with their basic counterpart for all standard arguments,
such as id/children/etc. This is to ensure they can all be directly dropped into existing Textual applications, before
any extension functionality is used. All Textual basic widgets should be 100% forward compatible with Textology widgets,
and all Textology widgets should be 100% backwards compatible with Textual widgets, in settings where no extended
functionality is being used.

This file must be kept in sync with the __init__.pyi's import list.
The pyi file is used by text editors and type checkers to see the lazily loaded classes.
"""

import typing
from importlib import import_module

from ._extensions import Widget

# Lazily load widgets to decrease startup time and allow multi version support.
if typing.TYPE_CHECKING:  # pragma: no cover. Exclude from coverage to allow lazy imports without import errors.
    from ._button import Button
    from ._button import SelectButton
    from ._extensions import Callback
    from ._extensions import Callbacks
    from ._extensions import Clickable
    from ._extensions import Static
    from ._extensions import ToggleButton
    from ._extensions import WidgetExtension
    from ._extensions import walk_all_children
    from ._horizontal_menus import HorizontalMenus
    from ._list_item import ListItem
    from ._list_item_header import ListItemHeader
    from ._list_item_meta import ListItemMeta
    from ._list_view import ListView
    from ._location import Location
    from ._modal_dialog import ModalDialog
    from ._multi_select import MultiSelect
    from ._page_container import PageContainer
    from ._popup_text import PopupText
    from ._store import Store
    from ._text import Text
    from ._textual._checkbox import Checkbox
    from ._textual._collapsible import Collapsible
    from ._textual._containers import Center
    from ._textual._containers import Container
    from ._textual._containers import Grid
    from ._textual._containers import Horizontal
    from ._textual._containers import HorizontalScroll
    from ._textual._containers import Middle
    from ._textual._containers import ScrollableContainer
    from ._textual._containers import Vertical
    from ._textual._containers import VerticalScroll
    from ._textual._content_switcher import ContentSwitcher
    from ._textual._data_table import DataTable
    from ._textual._digits import Digits
    from ._textual._directory_tree import DirectoryTree
    from ._textual._footer import Footer
    from ._textual._header import Header
    from ._textual._label import Label
    from ._textual._loading_indicator import LoadingIndicator
    from ._textual._log import Log
    from ._textual._markdown import Markdown
    from ._textual._markdown import MarkdownViewer
    from ._textual._option_list import OptionList
    from ._textual._pretty import Pretty
    from ._textual._progress_bar import ProgressBar
    from ._textual._radio_button import RadioButton
    from ._textual._radio_set import RadioSet
    from ._textual._rich_log import RichLog
    from ._textual._rule import Rule
    from ._textual._select import Select
    from ._textual._selection_list import SelectionList
    from ._textual._sparkline import Sparkline
    from ._textual._switch import Switch
    from ._textual._tabbed_content import TabbedContent
    from ._textual._tabbed_content import TabPane
    from ._textual._tabs import Tab
    from ._textual._tabs import Tabs
    from ._textual._text_area import TextArea
    from ._textual._text_input import TextInput
    from ._textual._tooltip import Tooltip
    from ._textual._tree import Tree
    from ._tree import LazyTree

_module_cache: dict[str, type[Widget]] = {
    "Widget": Widget,
}
_module_map = {
    "Button": "._button",
    "Callback": "._extensions",
    "Callbacks": "._extensions",
    "Center": "._textual._containers",
    "Checkbox": "._textual._checkbox",
    "Clickable": "._extensions",
    "Collapsible": "._textual._collapsible",
    "Container": "._textual._containers",
    "ContentSwitcher": "._textual._content_switcher",
    "DataTable": "._textual._data_table",
    "Digits": "._textual._digits",
    "DirectoryTree": "._textual._directory_tree",
    "Footer": "._textual._footer",
    "Grid": "._textual._containers",
    "Header": "._textual._header",
    "Horizontal": "._textual._containers",
    "HorizontalMenus": "._horizontal_menus",
    "HorizontalScroll": "._textual._containers",
    "LazyTree": "._tree",
    "Label": "._textual._label",
    "ListItem": "._list_item",
    "ListItemHeader": "._list_item_header",
    "ListItemMeta": "._list_item_meta",
    "ListView": "._list_view",
    "LoadingIndicator": "._textual._loading_indicator",
    "Location": "._location",
    "Log": "._textual._log",
    "Markdown": "._textual._markdown",
    "MarkdownViewer": "._textual._markdown",
    "Middle": "._textual._containers",
    "ModalDialog": "._modal_dialog",
    "MultiSelect": "._multi_select",
    "OptionList": "._textual._option_list",
    "PageContainer": "._page_container",
    "PopupText": "._popup_text",
    "Pretty": "._textual._pretty",
    "ProgressBar": "._textual._progress_bar",
    "RadioButton": "._textual._radio_button",
    "RadioSet": "._textual._radio_set",
    "RichLog": "._textual._rich_log",
    "Rule": "._textual._rule",
    "ScrollableContainer": "._textual._containers",
    "Select": "._textual._select",
    "SelectButton": "._button",
    "SelectionList": "._textual._selection_list",
    "Sparkline": "._textual._sparkline",
    "Static": "._extensions",
    "StaticFactory": "._extensions",
    "Store": "._store",
    "Switch": "._textual._switch",
    "Tab": "._textual._tabs",
    "TabbedContent": "._textual._tabbed_content",
    "Tabs": "._textual._tabs",
    "TabPane": "._textual._tabbed_content",
    "Text": "._text",
    "TextArea": "._textual._text_area",
    "TextInput": "._textual._text_input",
    "ToggleButtonFactory": "._extensions",
    "Tooltip": "._textual._tooltip",
    "Tree": "._textual._tree",
    "Vertical": "._textual._containers",
    "VerticalScroll": "._textual._containers",
    "WidgetExtension": "._extensions",
    "WidgetFactory": "._extensions",
    "walk_all_children": "._extensions",
}
__all__ = tuple(_module_cache.keys()) + tuple(_module_map.keys())


def __getattr__(attr_name: str) -> typing.Any:
    """Lazily load widgets, functions, and constants to decrease startup time."""
    try:
        return _module_cache[attr_name]
    except KeyError:
        pass

    if attr_name not in __all__:
        raise AttributeError(f"module 'textology.widgets' has no attribute '{attr_name}'")

    widget_module_path = _module_map.get(attr_name)
    module = import_module(widget_module_path, package="textology.widgets")
    attr = getattr(module, attr_name)
    _module_cache[attr_name] = attr
    return attr



"""Extended Rich text objects with additional customization."""

from __future__ import annotations

from typing import Literal
from typing import Optional

from rich.console import JustifyMethod
from rich.console import OverflowMethod
from rich.style import Style
from rich.text import DEFAULT_OVERFLOW
from rich.text import Span
from rich.text import Text as RichText
from rich.text import cell_len
from rich.text import set_cell_size
from typing_extensions import override

OverflowSide = Literal["left", "right"]


class Text(RichText):
    """Extended Rich text with standard color/style controls, and additional overflow controls."""

    __slots__ = [
        "_overflow_side",
        "_overflow_char",
    ]

    def __init__(
        self,
        text: str = "",
        style: str | Style = "",
        *,
        justify: JustifyMethod | None = None,
        overflow: OverflowMethod | None = None,
        overflow_side: OverflowSide = "right",
        overflow_char: OverflowSide = "…",
        no_wrap: bool | None = None,
        end: str = "\n",
        tab_size: Optional[int] = None,
        spans: list[Span] | None = None,
    ) -> None:
        """Initialize the text with style and overflow controls.

        Args:
            text: Default unstyled text.
            style: Base style for text.
            justify: Justify method as "left", "center", "full", "right".
            overflow: Overflow method: "crop", "fold", "ellipsis".
            overflow_side: Which side of the text to show ellipsis on as "left" or "right".
            overflow_char: Which character to use for the ellipsis.
            no_wrap: Disable text wrapping, or None for default.
            end: Character to end text with.
            tab_size: Number of spaces per tab, or ``None`` to use ``console.tab_size``.
            spans: A list of predefined style spans. Defaults to None.
        """
        super().__init__(
            text,
            style,
            justify=justify,
            overflow=overflow,
            no_wrap=no_wrap,
            end=end,
            tab_size=tab_size,
            spans=spans,
        )
        if len(overflow_char) > 1:
            raise ValueError("Overflow characters must be a single character")
        self._overflow_side = overflow_side
        self._overflow_char = overflow_char

    @override
    def blank_copy(self, plain: str = "") -> Text:
        copy_self = Text(
            plain,
            style=self.style,
            justify=self.justify,
            overflow=self.overflow,
            overflow_side=self._overflow_side,
            overflow_char=self._overflow_char,
            no_wrap=self.no_wrap,
            end=self.end,
            tab_size=self.tab_size,
        )
        return copy_self

    @override
    def copy(self) -> Text:
        copy_self = Text(
            self.plain,
            style=self.style,
            justify=self.justify,
            overflow=self.overflow,
            overflow_side=self._overflow_side,
            overflow_char=self._overflow_char,
            no_wrap=self.no_wrap,
            end=self.end,
            tab_size=self.tab_size,
        )
        copy_self._spans[:] = self._spans  # pylint: disable=protected-access
        return copy_self

    @override
    def truncate(
        self,
        max_width: int,
        *,
        overflow: OverflowMethod | None = None,
        pad: bool = False,
    ) -> None:
        # Matches parent truncate logic except to control side and char.
        _overflow = overflow or self.overflow or DEFAULT_OVERFLOW
        if _overflow != "ignore":
            length = cell_len(self.plain)
            if length > max_width:
                if _overflow == "ellipsis":
                    if self._overflow_side == "right":
                        self.plain = set_cell_size(self.plain, max_width - 1) + self._overflow_char
                    else:
                        self.plain = self._overflow_char + str(set_cell_size(self.plain[::-1], max_width - 1)[::-1])
                else:
                    self.plain = set_cell_size(self.plain, max_width)
            if pad and length < max_width:
                spaces = max_width - length
                self._text = [f"{self.plain}{' ' * spaces}"]
                self._length = len(self.plain)



"""Extended Textual, simple, clickable, button."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable

from rich.text import Text
from rich.text import TextType
from textual import events
from textual import widgets
from textual.message import Message
from textual.pad import HorizontalPad
from textual.reactive import reactive
from textual.widgets.button import ButtonVariant
from typing_extensions import override

from ._extensions import Callbacks
from ._extensions import Clickable
from ._extensions import WidgetExtension

if TYPE_CHECKING:
    from textual.app import RenderResult


class Button(widgets.Button, WidgetExtension, Clickable):
    """An extended, simple, clickable button."""

    def __init__(
        self,
        label: TextType | None = None,
        variant: ButtonVariant = "default",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize a Button widget with extension arguments.

        Args:
            label: The text that appears within the button.
            variant: The variant of the button.
            name: The name of the button.
            id: The ID of the button in the DOM.
            classes: The CSS classes of the button.
            disabled: Whether the button is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            label=label,
            variant=variant,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )

    async def intercept_message(self, message: Message) -> Message | None:
        """Update reactive attributes on press."""
        if isinstance(message, Button.Pressed):
            self.update_n_clicks()
        return message


class SelectButton(Button):
    """A button that tracks whether it is selected or not.

    Differs from RadioButton and ToggleButton in that it has all the same functionality as a Button,
    but also automatically updates and tracks whether it is currently "selected" or not on press events.
    """

    selected: bool = reactive(False, repaint=False, init=False)

    class Selected(Message):
        """Event sent when the button is selected/de-selected."""

        def __init__(self, button: SelectButton, selected: bool) -> None:
            """Initialize the selection update event.

            Args:
                button: The button that sent the event.
                selected: Whether the button is selected.
            """
            super().__init__()
            self.button = button
            self.selected = selected

        @property
        def control(self) -> SelectButton:
            """The button that was selected/deselected."""
            return self.button

    def __init__(
        self,
        label: TextType | None = None,
        *,
        selected: bool = False,
        selected_chars: str = "● ",
        deselected_chars: str = "⭘ ",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: Callbacks | None = None,
    ) -> None:
        """Initialize a SelectButton widget with extension arguments.

        Args:
            label: The text that appears within the button.
            selected: Whether the initial state of the button is selected.
            selected_chars: Character(s) to use as a label prefix when the item is selected.
            deselected_chars: Character(s) to use as a label prefix when the item is deselected.
            name: The name of the button.
            id: The ID of the button in the DOM.
            classes: The CSS classes of the button.
            disabled: Whether the button is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        self._selected_chars = Text.from_markup(selected_chars)
        self._deselected_chars = Text.from_markup(deselected_chars)
        super().__init__(
            label=label,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self.selected = selected

    async def intercept_message(self, message: Message) -> Message | None:
        """Update reactive attributes on press."""
        if isinstance(message, widgets.Button.Pressed):
            self.selected = not self.selected
        return message

    @override
    def render(self) -> RenderResult:
        label = (self._selected_chars if self.selected else self._deselected_chars) + self.label.copy()
        label.stylize_before(self.rich_style)
        return HorizontalPad(
            label,
            1,
            1,
            self.rich_style,
            self._get_rich_justify() or "center",
        )

    def watch_selected(self, selected: bool) -> None:
        """Monitor selected state to toggle indicator and trigger selection events."""
        self.post_message(self.Selected(self, selected))
        self.refresh()
